#include "SATmodel.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>


// NOTE: coeffs is the array of numFeat coefficients.  coeffs[numFeat] is the constant term.

SATmodel::SATmodel(double* satCoeffs, double* unsatCoeffs, double* cMean, double* cStd, int* featureID1, int* featureID2, double* classifyCoeff1, double* classifyCoeff2, int numFeat) {
   numFeatures = numFeat;
   c1 = satCoeffs;
   c2 = unsatCoeffs;
   ID1 = featureID1;
   ID2 = featureID2;
   mean = cMean;
   std = cStd;
   classCoeff1 = classifyCoeff1;
   classCoeff2 = classifyCoeff2;
}

SATmodel::SATmodel(double* Coeffs, double* cMean, double* cStd, int* featureID1, int* featureID2, int numFeat) {
   numFeatures = numFeat;
   c1 = Coeffs;
   ID1 = featureID1;
   ID2 = featureID2;
   mean = cMean;
   std = cStd;
   }

double SATmodel::estimateRunTime(double *features, double score1, double score2) {
   double val1 = c1[numFeatures];
   double val2 = c2[numFeatures];
   double val;
   double realFeatures;
   double pr1;
   double pr2;
   double maxCut, minCut;

   maxCut = log(DBL_MAX) -log(2);
   minCut = -log(DBL_MAX);

   pr1 = classCoeff1[0] + log(score1/score2)*classCoeff1[1];
   pr2 = classCoeff2[0] + log(score1/score2)*classCoeff2[1];
   
   for (int i=0; i<numFeatures; i++){
         int m = ID1[i];
         int n = ID2[i];
	 double x;
	 double y;
         if (m==0) 
		x = 1;
         else
		 x = features[m-1];
 	 
         if (n==0) 
		y = 1;
         else
		y = features[n-1];
         double tmpFeat = (x*y-mean[i])/std[i];

       val1 += c1[i] * tmpFeat;
       val2 += c2[i] * tmpFeat;
       pr1 += classCoeff1[i+2]*tmpFeat;
       pr2 += classCoeff2[i+2]*tmpFeat;
   }

   if (pr1 > maxCut) pr1=maxCut;
   if (pr1 < minCut) pr1=minCut; 
   if (pr2 > maxCut) pr2=maxCut;
   if (pr2 < minCut) pr2=minCut; 
   double pr = exp(pr1) /(exp(pr1)+exp(pr2));
   val = val1* pr + val2*(1-pr);
   return val;
}

double SATmodel::estimateRunTime(double *features) {
   double val1 = c1[numFeatures];
   double realFeatures;
   
   for (int i=0; i<numFeatures; i++){
         int m = ID1[i];
         int n = ID2[i];
	 double x;
	 double y;
         if (m==0) 
		x = 1;
         else
		 x = features[m-1];
 	 
         if (n==0) 
		y = 1;
         else
		y = features[n-1];
         double tmpFeat = (x*y-mean[i])/std[i];

       val1 += c1[i] * tmpFeat;
   }
   return val1;
}
