#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <iostream>
#include <unistd.h>
#include <fstream>


#include "SATinstance.h"
#include "SATpreprocess.h"
#include "SATclassification.h"
#include "SATmodel.h"
#include "BinSolver.h"

#include "global.h"

using namespace std;

Stopwatch gSW;
int gTimeOut;
// int numFeats=64;
int getTimeOut(void)
{
  char * value = getenv("SATTIMEOUT");

  if (value == NULL)
    return(2419200); // -- 4 weeks

  return atoi(value);
}

int main(int argc, char** argv)
{

  char *filename;
  char *preprocFilename;
  SATclassification* classifier;
  double thisAlgScore;

  long seed;
  char* strseed;

  int returnVal;
  SATinstance* prob;

  //   char *dir = dirname(argv[0]);

  if (argc < 2)
  {
    printf ("usage: satzilla <instance filename> [random seed]\n");
    exit(1);
  }
  else if(argc == 3)
  {
    seed = atol(argv[2]);
    strseed=argv[2];
  }
  else
  {
    /*
    seed = 42;
    strseed="42";
    */
    seed=(long)time(NULL);
    strseed = new char[64];
    sprintf(strseed, "%d", seed);
  }


  filename = argv[1];

  // -- start bookkeeping
  gTimeOut = getTimeOut();
  gSW.Start();
  BuildSolvers(strseed);


  printf("c run March_dl for 5 seconds ... \n");
  returnVal = binSolvers[3] -> execute(filename, 5);
  printf("c %s returned code %d.\n", binSolvers[3]->name, returnVal);
  if (returnVal == 10 | returnVal == 20)
  {
    fstream fin;
    fin.open(binSolvers[3]->outFileName);
    if(fin.is_open())
    {
      cout<<fin.rdbuf();
      printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
      return returnVal;
    }
  }



  //   -- First try localsearch
  printf("c run saps (UBCSAT implementation) for %d seconds ... \n", atoi(LS_PREP_TIMEOUT));
  returnVal = LSSolver -> execute(filename, -1);
  printf("c %s returned code %d.\n", LSSolver->name, returnVal);

  if (returnVal == 10)
  {
    fstream fin;
    fin.open(LSSolver->outFileName);
    if(fin.is_open())
    {
      cout<<fin.rdbuf();
      printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
      return returnVal;
    }
    // runLocalSearch(filename, strseed);
    // -- satisfied!
  }
  // == TODO: Clean up all preprocess/vline/solver files



  prob = new SATinstance(filename, seed);
  prob->ignoreBadFeats = true;

  printf("c satzilla computing base features\n");

  //   mt = FULL_MODEL;

  int featureOK = 1;
  prob->computeFeatures();

  //   int featNumber = prob->getNumFeatures();
  int featNumber = 33;
  double *rawFeat = prob->getFeatureVals();
  char** featnames= prob->getFeatureNames();
  for (int i=0; i<featNumber; i++)
  {
    //    cout<<rawFeat[i]<<endl;
    if (rawFeat[i] == -1)
    {
      cout<<"Error!!! " << featnames[i]<<endl;
      featureOK =0;
      break;
    }
  }

  if (featureOK)
  {
    printf("c satzilla doing unit-prop probe\n");
    if (prob->unitPropProbe(true) == DONE)
    {
      fstream fin;
      fin.open(prob->vlineFilename);
      if(fin.is_open())
      {
        cout<<fin.rdbuf();
        unlink(prob->vlineFilename);
        delete[] prob->vlineFilename;
        printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
        return 10;
      }
    }

    printf("c satzilla doing local search probe\n");
    if (prob->localSearchProbe(true) == DONE)
    {
      fstream fin;
      fin.open(prob->vlineFilename);
      if(fin.is_open())
      {
        cout<<fin.rdbuf();
        unlink(prob->vlineFilename);
        delete[] prob->vlineFilename;
        printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
        return 10;
      }
    }

    printf("c satzilla doing lobjois search-space size estimate\n");
    if (prob->lobjoisProbe(true) == DONE)
    {
      fstream fin;
      fin.open(prob->vlineFilename);
      if(fin.is_open())
      {
        cout<<fin.rdbuf();
        unlink(prob->vlineFilename);
        delete[] prob->vlineFilename;
        printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());
        return 10;
      }
    }

    //assert(numFeats == prob->getNumFeatures());
    //printf("c numFeats = %d, getNumFeatures = %d", numFeats, prob->getNumFeatures());


    // classification, compute classify sore of this instance

    int featNumber = prob->getNumFeatures();
    // int featNumber = 64;
    double *rawFeat = prob->getFeatureVals();
    char** featnames= prob->getFeatureNames();
//    cout << featNumber << endl;
    if( featNumber != 48)
    {
      cout << "Something wrong with the feature computation !!! " << endl;
      featureOK =0;
    }
    for (int i=0; i<featNumber; i++)
    {
//           cout<<featnames[i]<<endl;
      if (rawFeat[i] == -1)
      {
        cout<<"Error!!! " << featnames[i]<<endl;
        featureOK =0;
        break;
      }
    }

  }
  int bestAlgs[NUM_ALGS];
  double bestAlgScores[NUM_ALGS];
  double satScore;
  double unsatScore;
  if (featureOK==1)
  {
    switch (satzillaType)
    {
    case 1:
      classifier = new SATclassification(randomClassifyCoeff, featNumber, randomClassifyMean, randomClassifyStd);
      satScore = classifier->estimateClass(rawFeat);
      unsatScore = 1-satScore;
      //      cout << "sat score is:"<<satScore << endl;

      for (int i=0; i<NUM_ALGS; i++)
      {
        if (i==3 | i==4 | i==5 )
          thisAlgScore = (binSolvers[i]->models[satzillaType])->estimateRunTime(prob->getFeatureVals(), satScore, unsatScore);
        else
          thisAlgScore = 999.00;

        int j=0;
        while (j<i && thisAlgScore > bestAlgScores[j]) j++;
        for (int k=i; k>j; k--)
        {
          bestAlgs[k] = bestAlgs[k-1];
          bestAlgScores[k] = bestAlgScores[k-1];
        }
        bestAlgs[j]=i;
        bestAlgScores[j]=thisAlgScore;
      }
      break;
    case 2:

      classifier = new SATclassification(craftedClassifyCoeff, featNumber, craftedClassifyMean, craftedClassifyStd);
      satScore = classifier->estimateClass(rawFeat);
      unsatScore = 1-satScore;
      //      cout << "sat score is:"<<satScore << endl;
      for (int i=0; i<NUM_ALGS; i++)
      {
        if (i==2 | i==3 | i==5 | i==8 | i==6)
          thisAlgScore = (binSolvers[i]->models[satzillaType])->estimateRunTime(prob->getFeatureVals(), satScore, unsatScore);
        else
          thisAlgScore = 999.00;
        int j=0;
        while (j<i && thisAlgScore > bestAlgScores[j]) j++;
        for (int k=i; k>j; k--)
        {
          bestAlgs[k] = bestAlgs[k-1];
          bestAlgScores[k] = bestAlgScores[k-1];
        }
        bestAlgs[j]=i;
        bestAlgScores[j]=thisAlgScore;
      }
      break;
    default:
      for (int i=0; i<NUM_ALGS; i++)
      {
        if (i==0 | i==1 | i==2 | i==3 | i== 4)
          thisAlgScore = (binSolvers[i]->models[satzillaType])->estimateRunTime(prob->getFeatureVals());
        else
          thisAlgScore = 999.00;
        int j=0;
        while (j<i && thisAlgScore > bestAlgScores[j]) j++;
        for (int k=i; k>j; k--)
        {
          bestAlgs[k] = bestAlgs[k-1];
          bestAlgScores[k] = bestAlgScores[k-1];
        }
        bestAlgs[j]=i;
        bestAlgScores[j]=thisAlgScore;
      }
    }
    printf("c satzilla's ranking:\n");
    for (int i=0; i<NUM_ALGS; i++)
    {
      if (bestAlgs[i] >= NUM_ALGS/2)
        printf("c %d) %s+P: %f\n", i+1, binSolvers[bestAlgs[i]]->name, bestAlgScores[i]);
      else
        printf("c %d) %s: %f\n", i+1, binSolvers[bestAlgs[i]]->name, bestAlgScores[i]);
    }
  }
  else
  {
    printf("Satzilla is runing default solver ... \n ");
    for (int i=0; i<NUM_ALGS; i++)
      bestAlgs[i]=i;
  }
  int i=0;
  int preflag=0;
  while(true)
  {
    if(gSW.TotalLap() > gTimeOut)
    {
      printf("c Satzilla Timed out!\n");
      printf("s UNKNOWN\n");
      returnVal = 1;
      break;
    }

    if(i==NUM_ALGS)
    {
      printf("c ERROR: All Failed! \n");
      returnVal = 17;
      break;
    }
    // if it is a local search solver than set a cut off time for local search
    if (bestAlgs[i] >= NUM_ALGS/2)
      printf("c satzilla running %s with preprocessing ... \n", binSolvers[bestAlgs[i]]->name);
    else
      printf("c satzilla running %s ... \n", binSolvers[bestAlgs[i]]->name);
    fflush(stdout);

    //  returnVal = (*algos[bestAlgs[i]])(argcs[bestAlgs[i]], argvs[bestAlgs[i]]);

    if (preflag==0 & bestAlgs[i] >= NUM_ALGS/2)
    {
      printf("c Preprocessing...\n");
      preprocFilename = PreprocessCNF(filename);
      preflag=1;
    }
    if(bestAlgs[i] >= NUM_ALGS/2)
      returnVal = binSolvers[bestAlgs[i]] -> execute(preprocFilename, -1);
    else
      returnVal = binSolvers[bestAlgs[i]] -> execute(filename, -1);
    printf("c %s returned code %d.\n", binSolvers[bestAlgs[i]]->name, returnVal);

    if (returnVal == 10)
    {
      if (bestAlgs[i] >= NUM_ALGS/2)
      {
        bool vres=DoVLine(binSolvers[bestAlgs[i]]->outFileName, preprocFilename);
        binSolvers[bestAlgs[i]]->cleanup();
        if(vres)
        {
          break;
        }
      }
      else
      {
        fstream fin;
        fin.open(binSolvers[bestAlgs[i]]->outFileName);
        binSolvers[bestAlgs[i]]->cleanup();
        if(fin.is_open())
        {
          cout<<fin.rdbuf();
          break;
        }
      }
      printf("c ERROR in vline!, trying next\n");
      returnVal=-1;
      i++;
    }
    else if (returnVal == 20)
    {
      printf("s UNSATISFIABLE\n");
      binSolvers[bestAlgs[i]]->cleanup();
      i++;
      break;

    }
    else
    {
      printf("c %s returned code %d.\n", binSolvers[bestAlgs[i]]->name,returnVal);
      binSolvers[bestAlgs[i]]->cleanup();
      i++;
    }
  }

  printf("c SATZILLA_TIME=%lf\n", gSW.TotalLap());

  PreprocessorClean();

  return returnVal;
}
