#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "hypre/2clsEq.h"
#include "hypre/preprocessor.h"
#include "hypre/utilities.h"
#include "SATpreprocess.h"


enum state {TRUE, FALSE, EQV, NEWNUM};


namespace hypre {
  char *filename;
  char *outfile;
}

bool usingHypreFile;

void PreprocessorClean()
{
  if(usingHypreFile)
    unlink(hypre::outfile);
}

char *PreprocessCNF(char *filename)
{
  usingHypreFile = false;

  hypre::filename = filename;
  hypre::outfile = new char[512];
  sprintf(hypre::outfile, "%s", P_tmpdir);
  strcat(hypre::outfile, "/XXXXXX");
  hypre::outfile = mktemp(hypre::outfile);

  if(strlen(hypre::outfile) == 0)
    {
      fprintf(stderr, "c Couldn't create temp file\n");
      return(filename);
    }

  hypre::CONTRADICTION.val=false;
  if(!hypre::readProblem(filename)) {
    printf("c Problem reading CNF file.\n");
    //    return(NULL);
    return filename;
  }

  hypre::start_time = hypre::GetInternalRunTime();

  if(strchr(filename,'/'))
    filename=strrchr(filename, '/')+1;
  else if(strchr(filename,'\\'))
    filename=strrchr(filename, '\\')+1;
  hypre::preprocessor();

  if(hypre::GetInternalRunTime() - hypre::start_time > PREPRO_TIME_LIMIT) {
    delete[] hypre::outfile;
    //    return(NULL);
    return hypre::filename;
  }

  // This function's work is now performed in the beginning of cnfoutput().
  //gatherStats();
  if(!hypre::CONTRADICTION.val)
    hypre::cnfoutput();

  //temp
  //fprintReducedTheory();


  hypre::end_time = hypre::GetInternalRunTime();

  hypre::printSummary();
  hypre::recordSummary();

  usingHypreFile = true;
  return(hypre::outfile);
}



bool grepVLine(char* vline_filename)
{
  // -- go through the file, and return the vline
  FILE* vfp;

  printf("s SATISFIABLE\n");
  if ((vfp=fopen(vline_filename, "r"))==NULL)
    {
      printf("c ERROR: Could not open %s that contains output\n", vline_filename);
      return false;
    }

  char buf[1024];

  while( fgets(buf, 1024, vfp) )
    fputs(buf, stdout);

  fclose(vfp);

  return true;
}

bool DoVLine(char *vline_filename, char *cnf_filename)
{
  int i, retval, lit, orig_var, new_var;
  bool value;
  int OLDNVARS, NEWNVARS;
  FILE * vfp;
  FILE * cfp;
  char ch;
  char word[81];
  bool * assignments;
  int * varNumber;
  state * varState;
  int * newNum;

  if(!usingHypreFile)
    return grepVLine(vline_filename);

  //vline_filename = argv[1];
  if((vfp = fopen(vline_filename, "r")) == NULL) {
    printf("c ERROR: Could not open file named %s.\n", vline_filename);
    //usage();
    return false;
  }

  //cnf_filename = argv[2];
  if((cfp = fopen(cnf_filename, "r")) == NULL) {
    printf("c ERROR: Could not open file named %s.\n", cnf_filename);
    //usage();
    return false;
  }
  
  // Find the line in the cnf file that contains a the number of variables
  // in the original theory and the number of variables in the reduced
  // theory.  It begins with a "c *"
  while((retval = fscanf(cfp, "c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(cfp)) {
      printf("c ERROR: Could not find line with variable numbers in cnf file.\n");
      fclose(vfp);
      fclose(cfp);
      //exit(0);
      return false;
    }
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'))
      ;
  }

  // After the * are two numbers.  The first is the number of variable in
  // the original and the second is the number in the reduced theory.
  fscanf(cfp, "%d %d", &OLDNVARS, &NEWNVARS);

  // Skip the rest of the white space on this line.
  while(fscanf(cfp, "%c", &ch) && (ch != '\n'));

  // Set up data structures to store information about the number mapping.
  varNumber = new int[OLDNVARS+1];
  varState = new state[OLDNVARS+1];
  newNum = new int[OLDNVARS+1];
  assignments = new bool[NEWNVARS+1];


  // Now process the v-line file containing the variable truth assignment
  // for the reduced theory.

  // Find the line containing the variable assignments.  It starts with the
  // letter 'v'.
  while((retval = fscanf(vfp, "%c", &ch)) && (ch != 'v')) {
    if((retval == EOF) || feof(vfp)) {
      printf("c ERROR: Could not find the v-line in the file.\n");
      fclose(vfp);
      fclose(cfp);
      //exit(0);
      return false;
    }
    while(fscanf(vfp, "%c", &ch) && (ch != '\n'))
      ;
  }

  for(i = 1; i <= NEWNVARS; i++) {
    assignments[i] = false; //true;
  }
  char temp;
  while(fscanf(vfp, "%d", &lit) && (lit != 0)) {
    assignments[abs(lit)] = (lit > 0) ? true : false;
    while(true) {
      temp = fgetc(vfp);
      if(temp != ' ' && temp != '\n' && temp != 'v') {
        ungetc(temp, vfp);
        break;
      }
    }
  }
  
  
  // Create output v-line file which contains the matching truth assignments
  // using the variable numbering from the original cnf file.
  
  /*if((outfp = fopen("vline.out", "w")) == NULL) {
    printf("ERROR: Could not open file vline.out.\n");
    exit(1);
    }*/
  
  // Now, go back to looking through the cnf file and find and process
  // the
  
  // Find mapping of original cnf numbering to new cnf numbering inside
  // the new cnf file.
  while((retval = fscanf(cfp, "c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(cfp)) {
      printf("c ERROR: Could not find mapping in the cnf file.\n");
      fclose(vfp);
      fclose(cfp);
      //exit(0);
      return false;
    }
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'))
      ;
  }

  // Now begin to read the mapping from the old cnf file and store the
  // mapping information in two arrays.
  i = 1;
  while((retval = fscanf(cfp, " c %c", &ch)) && (ch != '*')) {
    if((retval == EOF) || feof(cfp)) {
      printf("c ERROR: cnf file ended prematurely.\n");
      fclose(vfp);
      fclose(cfp);
      //exit(0);
      return false;
    }

    // Skip lines with negative literals.
    while(fscanf(cfp, "%c", &ch) && (ch != '\n'));

    fscanf(cfp, "c %d %s", &orig_var, word);
    varNumber[i] = orig_var;

    //temp
    //fprintf(outfp, "%d ", orig_var);

    if(strcmp(word, "T") == 0)
      varState[i] = TRUE;
    else if(strcmp(word, "F") == 0)
      varState[i] = FALSE;
    else if(strcmp(word, "equivalent") == 0) {
      fscanf(cfp, " to %d", &new_var);
      varState[i] = EQV;
      newNum[i] = new_var;
    }
    // Arbitrarily assign variables whose truth value did not affect the
    // satisfiability of the theory to TRUE.
    else if(strcmp(word, "not") == 0) {
      varState[i] = TRUE;
      // Skip the rest of this line.
      while(fscanf(cfp, "%c", &ch) && (ch != '\n'));
    }
    else {
      if((new_var = strtol(word, NULL, 10)) == 0) {
        printf("c ERROR: Expected integer but did not find one.\n");
        //exit(1);
	return false;
      }
      varState[i] = NEWNUM;
      newNum[i] = new_var;
    }
    i++;
  }

  //temp
  //fprintf(outfp, "\n");

  // Now print out the v-line file for the old cnf's numbering of the
  // variables.
  fprintf(stdout, "s SATISFIABLE\n");
  fprintf(stdout, "v ");

  for(i = 1; i <= OLDNVARS; i++) {
    if(varState[i] == TRUE)
      fprintf(stdout, "%d ", varNumber[i]);
    else if(varState[i] == FALSE)
      fprintf(stdout, "-%d ", varNumber[i]);
    else if(varState[i] == EQV) {
      if(varState[abs(newNum[i])] == NEWNUM) {
        if(newNum[i] > 0)
          value = assignments[newNum[newNum[i]]];
        else
          value = !assignments[newNum[abs(newNum[i])]];

        if(value)
          fprintf(stdout, "%d ", varNumber[i]);
        else
          fprintf(stdout, "-%d ", varNumber[i]);
      }
      else if(((newNum[i] > 0) && (varState[newNum[i]] == TRUE))
              || ((newNum[i] < 0) && (varState[abs(newNum[i])] == FALSE)))
        fprintf(stdout, "%d ", varNumber[i]);
      else if(((newNum[i] < 0) && (varState[abs(newNum[i])] == TRUE))
              || ((newNum[i] > 0) && (varState[newNum[i]] == FALSE)))
        fprintf(stdout, "-%d ", varNumber[i]);
      else {
        printf("c ERROR: Ran into unexpected situation.\n");
        //exit(1);
	return false;
      }
    }
    else if(varState[i] == NEWNUM) {
      if(assignments[newNum[i]])
        fprintf(stdout, "%d ", varNumber[i]);
      else
        fprintf(stdout, "-%d ", varNumber[i]);
    }

#ifdef _DEBUG
    else
      printf("c ERROR: varState[%d] has an unknown value!\n", i);
#endif // _DEBUG

  }

  fprintf(stdout, "0\n");

  fclose(vfp);
  fclose(cfp);

  return true;
}
